/*
** minimal SSLProxy server for content switch project
** pieces of code taken from modssl which is protected under the below 
licence
**			ganesh kumar godavari (gkgodava@archie.uccs.edu)
*/


#define CLIENT_TIMEOUT 			30
#define SERVER_ROOT			"/home/ycai/caiyu"
#define LOG_FILE 			"log/ssl.log"
#define CA_FILE     			"testssl/ca/cacert.pem"
#define CA_PATH     			"testssl/ca"
#define KEY_FILE    			"testssl/private/private.key"
#define CERT_FILE   			"testssl/cert/newcert.pem"
#define RAND_FILE   			"testssl/random/random.pem" /* random is a junk file that contains any data--ensure it is not repeated*/
#define SERVER_PORT 			5111
#define SERVER_NAME			"128.198.61.51"


#define MAX_LINES   99
#define MAX_CHARS   4000
#define MAX_TXT     256
#define ITERATIONS  20
#define THRESHOLD   10.0


/* general libraries */
#include <setjmp.h>

#include "msock.h"
#include "misc.h"
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>


static SSL *ssl=(SSL *) NULL;

/* contants definition for access from path */
char *log_file;	
char *ca_file; 
char *ca_path;
char *key_file;
char *cert_file;
char *rand_file;

static int long_jumped=0,g_sock_fd;
static sigjmp_buf jmpbuf;
static volatile sig_atomic_t canjump;
static FILE *log_fp = (FILE *) NULL;

static int sslVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,int error,char *arg);
static void serveRequest(int fd);

void fatalError(char *format,...)
{
    va_list  args;

    va_start(args,format);
    (void) fprintf (stderr,"Error: ");
    vfprintf(stderr,format,args);
    va_end(args);
    exit(-1);
}

/*
** write all the junk!!  which u think may be useful to Administrator
**
*/
void logPrint(FILE * log_fp,char * format,...)
{
    va_list  args;
    time_t   now;

 
   char   *time_string;

    if(log_fp == NULL) {
	(void) fprintf(stderr,"unable to identify how the log file %s closed \n",LOG_FILE);
	return;
    }
	
    now=time(NULL);

    time_string=ctime(&now);
    time_string[strlen(time_string)-1]='\0';

    va_start(args,format);

    (void) fprintf(log_fp,"[%s]",time_string);
    (void) vfprintf(log_fp,format,args);

    va_end(args);
    fflush(log_fp);
    return;

}

static void sigAlarm(int sigtype)
{
    if (canjump == 0)
    {
        return;
    }
    canjump=1;
    long_jumped=1;
    siglongjmp(jmpbuf,1);
}

void make_absolute_path(void)
{
if (
(log_file = (char *) malloc (sizeof (SERVER_ROOT) + sizeof (LOG_FILE) + 1)) == NULL    ||
    (ca_file    = (char *) malloc( sizeof(SERVER_ROOT) + sizeof(CA_FILE) + 1)) == NULL ||
    (ca_path    = (char *) malloc( sizeof(SERVER_ROOT) + sizeof(CA_PATH) + 1)) == NULL ||
    (key_file   = (char *) malloc( sizeof(SERVER_ROOT) + sizeof(KEY_FILE) + 1)) == NULL ||
    (cert_file  = (char *) malloc( sizeof(SERVER_ROOT) + sizeof(CERT_FILE)+ 1)) == NULL ||
    (rand_file  = (char *) malloc( sizeof(SERVER_ROOT) + sizeof(RAND_FILE)+ 1)) == NULL )
  {
   (void) fprintf(stderr," unable to allocate memory for converting relative paths to absolute paths");
   exit(1);
  }
(void) sprintf(log_file,"%s/%s",SERVER_ROOT,LOG_FILE);
(void) sprintf(ca_file,"%s/%s",SERVER_ROOT,CA_FILE);
(void) sprintf(ca_path,"%s/%s",SERVER_ROOT,CA_PATH);
(void) sprintf(key_file,"%s/%s",SERVER_ROOT,KEY_FILE);
(void) sprintf(cert_file,"%s/%s",SERVER_ROOT,CERT_FILE);
(void) sprintf(rand_file,"%s/%s",SERVER_ROOT,RAND_FILE);
 return;
}

/*
** borrowed from mod_ssl
*/
int SSL_smart_shutdown(SSL *ssl)
{
    int i;
    int rc;

    /*
     * Repeat the calls, because SSL_shutdown internally dispatches through a
     * little state machine. Usually only one or two interation should be
     * needed, so we restrict the total number of restrictions in order to
     * avoid process hangs in case the client played bad with the socket
     * connection and OpenSSL cannot recognize it.
     */

    rc = 0;
    for (i = 0; i < 4 /* max 2x pending + 2x data = 4 */; i++) {
        if ((rc = SSL_shutdown(ssl)))
            break;
    }

    return rc;
}

int main(int argc, char **argv) 
{
    int  sock_fd,
         err,
	 rc,
         server_verify=SSL_VERIFY_NONE;

    char 
         szbuf[BUFSIZ],
         issuer[256];;

    SSL_CTX  *ctx;
    SSL_METHOD *method=SSLv23_server_method();

    X509 *ca_cert, *client_cert;

    FILE *ca_fp=(FILE *) NULL;
   
    SSL_load_error_strings();
    OpenSSL_add_ssl_algorithms();
    (void)make_absolute_path();     /* convert relative to absolute path */
    
    ctx=(SSL_CTX *) SSL_CTX_new(method);

    if (ctx == (SSL_CTX *) NULL)
    {
      fatalError("Unable to create new SSL CTX\n");
      goto ExitProcessing;
    }
     
    if (!SSL_CTX_load_verify_locations(ctx,ca_file,ca_path))
        fatalError("Failed in SSL_CTX_load_verify_locations()!\n");

    /*
    ** Load Private Key
    */
    err=SSL_CTX_use_RSAPrivateKey_file(ctx, key_file, SSL_FILETYPE_PEM);

    if (err == -1)
    {
       	(void) fprintf(stderr,"Error reading private key\n");
	SSL_CTX_free(ctx);
        exit(1);
    }

   /*
   ** Load Certificate
   */
    err=SSL_CTX_use_certificate_file(ctx, cert_file, SSL_FILETYPE_PEM);

    if (err == -1)
    {
        (void) fprintf(stderr,"Error reading certificate\n");
	SSL_CTX_free(ctx);
        exit(1);
    }

   /*
   ** check if the certificate and private key match
   */ 
   err = SSL_CTX_check_private_key(ctx);
 
    if (err == -1)
    {
       	(void) fprintf(stderr,"Error cerificate and private key donot match \n");
	SSL_CTX_free(ctx);
       	exit(1);
    }

    /*
    ** Load randomness
    */
    if (!RAND_load_file(rand_file,1024*1024))
    {
         (void) fprintf(stderr,"Unable to load Randomness for generating Entropy :-( \n");
    }

    /*
    ** read the ca certificate and save the issuer string, we'll compare
    ** the client's issuer with this one, if they match allow connection
    ** or zap him
    */
    ca_fp=fopen(ca_file,"r");
    if (ca_fp == (FILE *) NULL)
        fatalError("Failed to open Trusted CA certificate file: %s\n", ca_file);

    ca_cert=NULL;
    ca_cert=X509_new();
    if (!PEM_read_X509(ca_fp,&ca_cert,NULL,NULL))
        fatalError("Error reading trusted CA certificate fie: %s\n",ca_file);

     X509_NAME_oneline(X509_get_issuer_name(ca_cert),issuer,256);
    if (issuer == (char *) NULL)
        fatalError("No issuer for trusted CA certificate file!\n");


    if (ca_cert != NULL)
        X509_free(ca_cert);

   (void) fclose(ca_fp);

#if 1
     /* verifying client --- set parameter here */

     server_verify=SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT|SSL_VERIFY_CLIENT_ONCE;

     (void) SSL_CTX_set_verify(ctx,server_verify,NULL);
#endif


    /* open the log file */
    if((log_fp=fopen(log_file,"a+"))== NULL) 
	(void) fprintf(stderr, "unable to open the log file");
 
    /*
    ** most cool Unix will take care of Zombies if we ignore SIGCHLD
    */
    signal(SIGCHLD,SIG_IGN);
    signal(SIGPIPE,SIG_IGN);
    signal(SIGALRM,sigAlarm);

    noCoredump(); 
    daemonize(); 

    /*
    ** open the server socket
    */
    sock_fd= serverSocket((u_short) SERVER_PORT ,1000);

    if (sock_fd == -1) {
        (void) fprintf(stderr,"Failed to create server\n");
        exit(1);
    }
    g_sock_fd=sock_fd;

    alarm(CLIENT_TIMEOUT);

    sigsetjmp(jmpbuf,1);
    if (long_jumped == 1)
    {
        /*
        ** back from a long jump, alarm timer expired
        */
        long_jumped=0;
        canjump=1;
        goto ExitProcessing;
    }

     ssl=SSL_new(ctx);
     if (ssl == NULL)
         fatalError("Failed in SSL_new()!\n");

     SSL_set_fd(ssl,sock_fd);

    /*
    ** cleanup the structures 
    */
    SSL_clear(ssl);
    SSL_set_session(ssl,NULL);
    SSL_set_accept_state(ssl);

    err=SSL_accept(ssl);
   if (SSL_get_error(ssl, err) == SSL_ERROR_ZERO_RETURN)
      {
         /*
         ** case where the connection was closed before any data was transferred
         */
          (void) fprintf(stderr, "SSL handshake stopped: connection was closed");
  	  goto ExitProcessing;
          return 0;
       }
     else if (ERR_GET_REASON(ERR_peek_error()) == SSL_R_HTTP_REQUEST)
          {
                /*
                ** case where OpenSSL has recognized a HTTP request => client speaks plain HTTP on our HTTPS
                */
		(void)fprintf(stderr, "SSL handshake failed: HTTP spoken on HTTPS port; ");
		goto ExitProcessing;
                return 0;
          }
     else if (SSL_get_error(ssl, err) == SSL_ERROR_SYSCALL)
          {
                if (errno > 0)
                    (void) fprintf(stderr,"SSL handshake interrupted by system "
                            "[Hint: Stop button pressed in browser?!]\n");
                else

                    (void) fprintf(stderr, "Spurious SSL handshake interrupt"
                            "[Hint: Usually just one of those OpenSSL confusions!?]\n");

		goto ExitProcessing;
		return 0;
          }
	else if (err == -1)
              {
	        fatalError("Error : unknown error in SSL_accept()\n");
		goto ExitProcessing;
		return 0;
	      }

#if 1
     if(SSL_get_verify_result(ssl) != X509_V_OK)
         fatalError("Client certificate does not verify\n");

     client_cert=SSL_get_peer_certificate(ssl);

     if (client_cert == NULL)
         fatalError("Client does not  have a certificate\n");
#endif
    serveRequest(sock_fd);

ExitProcessing:
    close(sock_fd);
    SSL_set_shutdown(ssl,SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
    SSL_smart_shutdown(ssl);
    SSL_CTX_free(ctx);
    SSL_free(ssl);
    exit(1);

}


void demoResponse(char *url)
{
 char body[1024];
 char content_length[64];
 sprintf(body," <!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">	\
                   <HTML><HEAD>						\
                   <TITLE>SCID</TITLE>				\
                   </HEAD><BODY>					\
                   <H1>Not Found</H1>					\
                   The requested URL %s was not found on this server.<P>	\
                   <HR>							\
                   <ADDRESS>Apache/1.3.14 Server at %s  Port 80</ADDRESS>	\
                   </BODY></HTML>",url,SERVER_NAME);

 sprintf(content_length,"Content-Length: %d\r\n",strlen(body));

 sockPutsSSL(ssl,"HTTP/1.0 404 OK\r\n");
 sockPutsSSL(ssl,"MIME-version: 1.0\r\n");
 sockPutsSSL(ssl,content_length);
 sockPutsSSL(ssl,"Content-Type: text/html\r\n\r\n");


 sockPutsSSL(ssl,body);

 return;
}

/*
** cai yu --------------- your work goes here
*/
static void serveRequest(int fd)
{
/*     demoResponse("hello testing .....");*/
//	execl("/bin/ls","ls","-l",0);
	getLoad();

  return;
}

int run_popen(char **lineptr, char *buffer, char *exe_str)
 {
   int i, cnt=0;
   FILE *ptr, *popen();
   char *bufstart, *bufend, line[MAX_TXT], *fgets();

   bufstart = buffer;
   bufend = buffer + MAX_CHARS;
   if((ptr =   popen(exe_str, "r")) == NULL)
       (void) fprintf(stderr,"Couldn't open pipe");

   for(i=0; i<MAX_LINES; i++)
       {
       if(fgets(line, MAX_TXT, ptr) == NULL)
          break;
       line[strlen(line)-1] = '\0';
       if((bufstart + strlen(line) + 1) >= bufend)
          (void) fprintf(stderr, "Line too long");
       lineptr[i]=bufstart;
       strcpy(bufstart,line);
       bufstart += strlen(line)+1;
       cnt++;
       }

  fclose(ptr);
  return cnt;
 }

int getLoad()
{
 char *slines[MAX_LINES];
 char sbuf[MAX_CHARS];
 char tmp[500];

 char *client_ip = "128.198.61.51";
 char *client_gw = "128.198.61.1";
 char *proxy_ip = "128.198.60.42 ";
 char *proxy_gw = "128.198.60.1 ";
 char *gw_ip = "128.198.60.200";
 char *gw_gw = "128.198.60.129";
 char *target_ip = "128.198.60.201";
 char *tunl = "tunl1";
 char *bcom_str = "sh /home/ycai/caiyu/ipip.sh";

 (void) sprintf(tmp,"%s %s %s %s %s %s %s %s %s", bcom_str,client_ip,client_gw,proxy_ip,proxy_gw,gw_ip,gw_gw,target_ip,tunl);
 bcom_str = tmp;
 printf("%s\n",bcom_str);
 int count, i;

 count=run_popen(slines, sbuf, bcom_str);
 if (count < 1)
    return -1;

 for(i=0; i< count; i++)
         printf (" line %d :: %s\n",i, slines[i]);

 return 1;
}



/*---------------------------------------------------------------------*/

static int sslVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,int error, 
char *arg)
{
    char
        *cip, *issuer_ca, s[64];

    X509_NAME_oneline(X509_get_subject_name(xs),s,64);
    if (s != (char *) NULL)
    {
        if (ok)
        {
            (void) fprintf(stderr,"ok depth=%d,subject=%s\n",depth,s);
        }
        else
            (void) fprintf (stderr,"depth=%d error=%d,subject=%s\n",
                            depth,error,s);
          free(s); 
    }
    X509_NAME_oneline(X509_get_issuer_name(xs),s,64);
    if (s == (char *) NULL)
    {
        (void) fprintf (stderr,"verify error\n");
        ok=0;
    }

    (void) fprintf(stderr,"Issuer=%s\n",s);
    ok=1;
    return (ok);
}
