/* Attempt to send an ICMP packet, writing the ICMP and IP headers. 
 * (Lacks adding the time to the end of the packet).
 * Thamer Al-Herbish shadows@whitefang.com
 */

#include <stdlib.h>
#include <stdio.h>

#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>

#include <strings.h>
#include <unistd.h>
#include <netdb.h>



#if !defined(IPVERSION)
#define IPVERSION 4  /* Incase some system does not have this definition. */
#endif               /* We'll always be using 4 as the version anyway. */

#define BUFFSIZE (sizeof(struct icmphdr) + sizeof(struct iphdr))


unsigned int hostid;
int     max_ttl = 30;
int     nprobes = 3;
struct sockaddr_in dest;
int ident;



int in_cksum(ptr, nbytes)
     register u_short        *ptr;
     register int            nbytes;
{
  register long           sum;            /* assumes long == 32 bits */
  u_short                 oddbyte;
  register u_short        answer;         /* assumes u_short == 16 bits */

  /*
   * Our algorithm is simple, using a 32-bit accumulator (sum),
   * we add sequential 16-bit words to it, and at the end, fold back
   * all the carry bits from the top 16 bits into the lower 16 bits.
   */

  sum = 0;
  while (nbytes > 1)  {
    sum += *ptr++;
    nbytes -= 2;
  }

  /* mop up an odd byte, if necessary */
  if (nbytes == 1) {
    oddbyte = 0;            /* make sure top half is zero */
    *((u_char *) &oddbyte) = *(u_char *)ptr;   /* one byte only */
    sum += oddbyte;
  }

  /*
   * Add back carry outs from top 16 bits to low 16 bits.
   */

  sum  = (sum >> 16) + (sum & 0xffff);    /* add high-16 to low-16 */
  sum += (sum >> 16);                     /* add carry */
  answer = ~sum;          /* ones-complement, then truncate to 16 bits */
  return(answer);
}

/*
** Return address in network byte order given the host address.
*/
unsigned int createConnection(char *hostname)
{

  struct hostent *he;
  int on = 1;
  int sockfd;

  bzero((char *)&dest,sizeof(dest));
  dest.sin_family = AF_INET;
  
  /*
  ** Lookup hosttable from the DNS.
  */
  he = gethostbyname(hostname);

  if (he)
    memcpy((caddr_t)&dest.sin_addr.s_addr, he->h_addr, he->h_length);
  else
    return -1;

  if (dest.sin_addr.s_addr == -1)
   {
    (void) fprintf(stderr, " unknown hostname %s \n", hostname);
    return -1;
   }


  if((sockfd = socket(AF_INET,SOCK_RAW,IPPROTO_RAW)) < 0)  {
    perror("socket");
    return (-1);
  }
  
  if(setsockopt(sockfd,IPPROTO_IP,IP_HDRINCL,(char *)&on,sizeof(on)) < 0)  {
    perror("setsockopt");
    return (-1);
  }

 return sockfd;

}



int sendEchoPacket(int sockd, int ttl)
{
  struct icmphdr *icmphdr;
  struct iphdr *iphdr;
  unsigned char buff[BUFFSIZE];
 
  iphdr = (struct iphdr *)buff;
  ident = (getpid() & 0xffff) | 0x8000 ;

  bzero((char *)iphdr,sizeof(struct iphdr));

  iphdr->ihl = ((u_char *) (iphdr + 1) - (u_char *)iphdr) >> 2;
  iphdr->version = IPVERSION;
    
#ifdef IP_LEN_HORDER
  iphdr->tot_len = (sizeof(struct iphdr) + sizeof(struct icmphdr));
#else
  iphdr->tot_len = htons(sizeof(struct iphdr) + sizeof(struct icmphdr));
#endif
  
  iphdr->id = htons(getpid());
  iphdr->ttl = ttl;
  iphdr->protocol = IPPROTO_ICMP;

  iphdr->daddr = dest.sin_addr.s_addr;

  iphdr->check = in_cksum((unsigned short *)iphdr,sizeof(struct iphdr));
  
  icmphdr = (struct icmphdr *)(buff +sizeof(struct iphdr));

  bzero((char *)icmphdr,sizeof(struct icmphdr));
  
  icmphdr->type = ICMP_ECHO;
  icmphdr->un.echo.sequence = getpid();
  icmphdr->un.echo.id =  getpid();
  
  icmphdr->checksum = in_cksum((unsigned short *)icmphdr, sizeof(struct icmphdr));

  if(sendto(sockd,(char *)buff,sizeof(buff),0x0, (struct sockaddr *)&dest,sizeof(dest)) < 0)  {
    perror("sendto");
    return -1;
  }

 return 1;
  
}


/*
** check if some data was send to us hopefully our icmp packet
*/
int waitforreply(int sockfd)
{
  int maxfd;
  int err;

  struct timeval  tv;
  fd_set   readfds;

  FD_ZERO(&readfds);
  maxfd=sockfd+1;
  FD_SET(sockfd,&readfds);

  tv.tv_sec=3;
  tv.tv_usec=0;

  err= select(maxfd,&readfds,NULL,NULL,&tv) ;
  if (err == 0)  return 0;

 if(FD_ISSET(sockfd,&readfds))
   return 1;

 return 0;
}


int recvEchoPacket(int sockd)
{
  struct icmphdr *icmphdr;
  struct iphdr *iphdr;
  unsigned char buff[BUFFSIZE];
  int iphdrlen;
  int nRet;
  int flen;


  flen = sizeof(dest);
  
  if((nRet = recvfrom(sockd,(char *)buff,sizeof(buff),0x0, (struct sockaddr *)&dest,&flen)) < 0)  {
    perror("recvfrom");
    return -1;
  }

 
  iphdr = (struct iphdr *)buff;

  if (iphdr->id != htons(getpid()) ) return -1;

  if (iphdr->protocol != IPPROTO_ICMP) return -1;

  iphdrlen = iphdr->ihl << 2;

  icmphdr = (struct icmphdr *)(buff+iphdrlen);

  printf("%d  ", iphdr->saddr);

  

 return 1;
  
}


int tracer(char *host)
{
  int sockd;

  sockd = createConnection(host);

  sendEchoPacket(sockd, 1);
#if 0
 if ( waitforreply(sockd) == 0)
 {
  (void) fprintf(stderr, "\nTimeout");
 }
else
#endif
  recvEchoPacket(sockd);

return 1;

}

int main(int argc,char *argv[])
{
  struct icmphdr *icmphdr;
  struct iphdr *iphdr;
  unsigned char buff[BUFFSIZE];
 
 if(argc < 2) {
   fprintf(stderr,"usage: %s source-ip dest-ip\n",argv[0]);
   exit(-1);
 }

return tracer(argv[1]);
  
 exit(0);
}




