/* THIS SOFTWARE IS PROVIDED BY RICE UNIVERSITY AND */
/* CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED */
/* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR */
/* PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL RICE UNIVERSITY */
/* OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, */
/* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES */
/* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE */
/* GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR */
/* BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF */
/* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT */
/* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT */
/* OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE */
/* POSSIBILITY OF SUCH DAMAGE. */

/* sclient.c v2.0, author: Gaurav Banga gaurav@cs.rice.edu */

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <string.h>
#include <sys/time.h>
#include <netinet/tcp.h>	
#include <fcntl.h>
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>

#ifndef __osf__
#include <machine/cpufunc.h>
int TRUE = 1;
int FALSE = 0;
#else
typedef unsigned long quad_t;
quad_t rdtsc();
#endif


char *host_machine = "appenzell.cs.rice.edu";
char *target_file = "index.html";
char *port_string = "8080";
char *reps = "100";
char *num_clients_string = "1";
char *rate_string = "200";
char *mhz_string = "400";

int global_port_num;
char *global_host_machine;

struct timeval start_time, end_time, zero_time;

typedef enum states_t {
  ST_NOTHING, ST_WAITING, ST_READING, ST_WRITING, ST_NUMSTATES
} states_t;

typedef struct {
  int fd;
  states_t state;
  int bytes_read;
  int partial_req_written;
  int num_reqs_done;
  unsigned int timestamp;
  struct timeval start_time;
} client_t;

#define PERCONBUFSIZE 6000000

unsigned last_connect;
int num_clients;
client_t all_clients[4000];
int num_reps_left;
int reqs_done;
long total_bytes_read;
fd_set master_read_set;
fd_set master_write_set;
int rate;
int cpu_mhz;
unsigned cycles_per_request;
unsigned max_connect_cycles;

void finish(void);
long diff_time(struct timeval *a, struct timeval *b);

char *request_string;
int request_string_len;

unsigned sel_before, sel_after;
unsigned long miss_sum, n_samples;
unsigned long sel_sum, n_sel_samples;
unsigned long diff_sum, n_diff_samples;

#define DIFFCYCLES(a, b)   (((a) > (b))? ((a) - (b)):(0xFFFFFFFF - (b) + (a)))


int deadline_passed(void)
{
  unsigned now = (unsigned int)(rdtsc() & 0xFFFFFFFF);
  unsigned age = DIFFCYCLES(now, last_connect);

  if(age < cycles_per_request)
    return 0;
  else {
    n_samples++;
    miss_sum += (unsigned long)(age - cycles_per_request);
    return 1;
  }
}


int too_long_connect(client_t *cl)
{
  unsigned now = (unsigned int)(rdtsc() & 0xFFFFFFFF);
  unsigned age = DIFFCYCLES(now, cl->timestamp);

    
  if(age < max_connect_cycles)
    return 0; 
  else 
    return 1;
}

void print_stats(void)
{
  double speed;
  double diff;
  int reps_done;

  gettimeofday(&end_time, 0);
  printf("Elapsed time: %.2f microseconds\n",
	 (((end_time.tv_sec - start_time.tv_sec) * 1000000.0) 
	  + (end_time.tv_usec - start_time.tv_usec)));

  diff = diff_time(&start_time, &end_time);
  if(diff == 0)
    diff = 1.0;

  speed = 8.0 * total_bytes_read/diff;
  reps_done = atoi(reps) - num_reps_left;
  printf("reps done: %d\n", reps_done);
  printf("request rate: %.1lf\n", reqs_done/((diff/1e6)));
  printf("[ %.3lf Mb/s, %s %.3f S, %ld B, %d R, %d cl, pt %d %.1lf c/s]\n",
	 speed, target_file, diff/1e6, total_bytes_read, 
	 reps_done, num_clients, atoi(port_string),
	 reps_done/(diff/1e6));

  printf("average response time: %.1lf us\n", diff_sum/(1.0 * n_diff_samples));
  printf("maximum number of pseudo clients: %d\n", num_clients);
}

static struct hostent *ent;
static struct sockaddr_in soc_address;
static struct sockaddr_in *sin_ptr = &soc_address;

void init_clients(char *hostname, int port)
{
  struct in_addr *ip_addr;

  if(hostname) {
    ent = gethostbyname(hostname);
    if(ent == NULL) {
      fprintf(stderr, "lookup on server's name failed\n");
      exit(1);
    }  
    ip_addr = (struct in_addr *)(*(ent->h_addr_list));
    sin_ptr->sin_family = AF_INET;
    bcopy(ip_addr, &(sin_ptr->sin_addr), sizeof(struct in_addr));
  }

  if(!ent) {
    if (hostname)
      printf("error - didn't get host info for %s\n", hostname);
    else
      printf("error - never called gethostbyname\n");
    finish();
  }

  if(port)
    sin_ptr->sin_port = htons(port);
}


static void 
change_state(client_t *client, states_t new_state)
{

  switch(client->state) {
  case ST_READING:
    FD_CLR(client->fd, &master_read_set);
    break;
  case ST_WRITING:
    FD_CLR(client->fd, &master_write_set);
    break;
  }

  client->state = new_state;

  switch(new_state) {		
  case ST_READING:
    FD_SET(client->fd, &master_read_set);
    break;
  case ST_WRITING:
    FD_SET(client->fd, &master_write_set);
    break;
  }
}


void simple_connect(int port, char *hostname)
{
  int status;
  int sendwin = 64*1024;
  int rcvwin = 64*1024;
  int fd;
  client_t *client;

  fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if(fd == -1) {
    perror("socket");
    finish();
  }

  client = &all_clients[fd];
  bzero(&all_clients[fd], sizeof(client_t));
  client->fd = fd;
  if(fd > (num_clients - 1))
    num_clients++; 

  if(fcntl(client->fd, F_SETFL, O_NDELAY) < 0)
    perror("fcntl");

  client->timestamp  = (unsigned int)(rdtsc() & 0xFFFFFFFF);
  gettimeofday(&(client->start_time), 0);
  last_connect = client->timestamp;

  status = connect(client->fd, (struct sockaddr *)&soc_address, 
		   sizeof(soc_address));


  if((status < 0) && (errno != EINPROGRESS)) {
    printf("errno is %d\n", errno);
    perror("connect");
    finish();
  }
  reqs_done++;
  change_state(client, ST_WRITING);
  
  if(setsockopt(client->fd, SOL_SOCKET, SO_SNDBUF,
		&sendwin, sizeof(sendwin)) < 0) {
    printf("%d\n", client->fd);
    perror("setsockopt");
  }

  if(setsockopt(client->fd, SOL_SOCKET, SO_RCVBUF,
		&rcvwin, sizeof(rcvwin)) < 0) {
    printf("%d\n", client->fd); 
    perror("setsockopt");
  }
}

void abort_connection(client_t *client);

int send_http(client_t *client)
{
  int res;
  int bytes_written;
  
  client->bytes_read = 0;

  bytes_written = client->partial_req_written;

  if(bytes_written) {
    res = write(client->fd, &request_string[bytes_written], 
		(request_string_len - bytes_written));

    if(res > 0)
      res += bytes_written;
  } else
    res = write(client->fd, request_string, request_string_len);

  if(res <= 0) {
    if(errno == EAGAIN)
      return(TRUE);
    abort_connection(client);
  }

  if(res != request_string_len)  {
    client->partial_req_written = res;
    return(TRUE);
  }

  if(bytes_written)
    client->partial_req_written = 0;

  change_state(client, ST_READING);
  return(FALSE);
}

void finish(void) 
{
  print_stats();
  exit(-1);
}



void done_with_connection(client_t *client)
{
  struct timeval now;
  long diff;

  total_bytes_read += client->bytes_read;
  client->num_reqs_done++;
  close(client->fd);
  gettimeofday(&now, 0);

  diff = diff_time(&(client->start_time), &now);
  diff_sum += (unsigned long)diff;
  n_diff_samples++;

  num_reps_left--;
  if(num_reps_left > 0) {
    if(deadline_passed()) {
      change_state(client, ST_NOTHING);
      simple_connect(0, 0); 
    } else
      change_state(client, ST_WAITING);
  } else
    change_state(client, ST_NOTHING);
}

void abort_connection(client_t *client)
{
  close(client->fd);
  if(num_reps_left > 0) {
    change_state(client, ST_NOTHING);
    simple_connect(0, 0); 
  } else
    change_state(client, ST_NOTHING);
}

#define READINCREMENTS 8*1024
char single_read_buf[2*READINCREMENTS]; 

void read_from_connection(client_t *client)
{
  int status;

  if(client->bytes_read + READINCREMENTS > PERCONBUFSIZE) {
    printf("PERCONBUFSIZE too small\n");
    finish();
  }

  status = read(client->fd, single_read_buf, READINCREMENTS); 
  
  if((status < 0) && (errno != EAGAIN)) {
    perror("read");
    finish();
  } 
    
  client->bytes_read += status;
  
  if (status == 0) {
    done_with_connection(client);
    return;
  }
}

char req_buf[1024];

char *build_http_request(char *filename)
{
#define CR '\015'
#define LF '\012'
  char *start;
  char *res;

  sprintf(req_buf, "GET http://fxp2.cs.rice.edu:8080/%s HTTP/1.0%c%c", filename, CR, LF);

  start = &req_buf[strlen(req_buf)];
  sprintf(start, "Accept: text/plain%c%c", CR, LF);
  start = &req_buf[strlen(req_buf)];
  sprintf(start, "Accept: text/html%c%c", CR, LF);
  start = &req_buf[strlen(req_buf)];
  sprintf(start, "Accept: */*%c%c", CR, LF);
  start = &req_buf[strlen(req_buf)];
  sprintf(start, "%c%c", CR, LF);
  return(&req_buf[0]);
}


long diff_time(struct timeval *a, struct timeval *b)
{
  return((b->tv_sec - a->tv_sec)*1000000L +
	 (b->tv_usec - a->tv_usec));
}


void signal_handler(sig, code, scp, addr)
     int sig, code;
     struct sigcontext *scp;
     char *addr;
{
  finish();
  exit(-1);
}


struct {
  char *name;
  char **value_pointer;
} string_options[] = {
  {"-host", &host_machine},
  {"-file", &target_file},
  {"-port", &port_string},
  {"-reps", &reps},
  {"-clients", &num_clients_string},
  {"-rate", &rate_string},
  {"-mhz", &mhz_string},
  {0,0}
};

void main(int argc, char *argv[])
{
  int sock;
  int i;
  int port;
  
  for(i = 1; i < argc; i++) {
    int j;
    for(j = 0; string_options[j].name; j++) {
      if(strcmp(string_options[j].name, argv[i]) == 0) {
	*string_options[j].value_pointer = argv[i+1];
	if(i+1 >= argc) {
	  printf("no value supplied for %s option\n", 
		 string_options[j].name);
	  exit(-1);
	}
	i++; /* skip over value for string switch */
	break;
      }
    }
  }
	
  for(i = 0; string_options[i].name; i++)
    printf("%s %s\n", string_options[i].name, 
	   *string_options[i].value_pointer);
  
  port = atoi(port_string);
  cpu_mhz = atoi(mhz_string);
  rate = atoi(rate_string);

  request_string = build_http_request(target_file);
  request_string_len = strlen(request_string);

  num_clients = atoi(num_clients_string);
  if(num_clients < 1) {
    printf("too few clients\n");
    exit(-1);
  }
	
  cycles_per_request = 
    (unsigned int)(cpu_mhz * (1000000.0/rate));	


  max_connect_cycles = (50000 * cpu_mhz);

  printf("A request will be issued every: %d cycles i.e. every %d us\n",
	cycles_per_request, (cycles_per_request/cpu_mhz));

  num_reps_left = atoi(reps);
  if(num_reps_left < 1) {
    printf("too few reps\n");
    exit(-1);
  }

  init_clients(host_machine, port);

  global_port_num = port;
  global_host_machine = host_machine;
  
  gettimeofday(&start_time, 0);
  
  signal(SIGPIPE, SIG_IGN);
  signal(SIGINT, signal_handler);
  
  for(i = 0; i < 32; i++) {
    simple_connect(port, host_machine);
    port = 0;
    host_machine = NULL;
  }

  while(num_reps_left > 0) {
    fd_set readfds;
    fd_set writefds;
    int num_ready;
    bcopy(&master_read_set, &readfds, sizeof(readfds));
    bcopy(&master_write_set, &writefds, sizeof(writefds));

    num_ready = select(FD_SETSIZE, &readfds, &writefds, NULL, &zero_time);

    if(num_ready < 0) {
      perror("select");
      finish();
    }
    
    for(i = 0;  i < num_clients; i++) {
      states_t st;
      st = all_clients[i].state;
      if((st == ST_READING)
	 && (FD_ISSET(all_clients[i].fd, &readfds)))
	read_from_connection(&all_clients[i]);
      else if(st == ST_WRITING) {
	if(FD_ISSET(all_clients[i].fd, &writefds)) {
	  send_http(&all_clients[i]);
	} 
          else if(too_long_connect(&all_clients[i])) {
	  abort_connection(&all_clients[i]);
	}
      } else if(st == ST_WAITING) {
	/* has the deadline passed? */
	if(deadline_passed()) 
	  simple_connect(0, 0);
      }
      if(deadline_passed()) {
	simple_connect(0, 0);
      }
    }
  }
  print_stats();
}

