/*
**
** i have written this code to understand how ICMP Protocol works
** this program can be run simultaniously by many users.
**
** simple Ping Utility written by Ganesh Godavari
**                               gkgodava@cs.uccs.edu
*/


/*
** working of the code
**
** send an icmp packet wait for the response, calculate the Round Trip Time and display it
*/

#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <setjmp.h>

#include <sys/time.h>
#include <sys/socket.h>
#include <sys/file.h>

#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netdb.h>

#include "ping.h"

int transmited = 0; /* # of packets transmited */
int received =0;    /* # of packets received */

struct sockaddr_in dest;

pid_t ident;		/* to identify that the icmp packet is for us only */

/*
** variables to display the statistics of the Round Trip Times
*/

float tmin = 999999999.999999999;
float tmax;
float tsum;


int debug;
int packetsize = 32;  /* default size of the icmp packet */
int packetcount = 8;  /* default no of packets to be transmitted */


void help_msg(char *name) {
  printf("usage: %s [<host address>] [ -d -h <hostname> -s <packetsize> -p <# of packets> ] \n",name);
  printf("-d turn on the debugging \n");
  
  return;
}

/*
** packet check sum program written in c
** i borrowed it from stevens book
*/

int in_cksum(ptr, nbytes)
     register u_short        *ptr;
     register int            nbytes;
{
  register long           sum;            /* assumes long == 32 bits */
  u_short                 oddbyte;
  register u_short        answer;         /* assumes u_short == 16 bits */
  
  /*
   * Our algorithm is simple, using a 32-bit accumulator (sum),
   * we add sequential 16-bit words to it, and at the end, fold back
   * all the carry bits from the top 16 bits into the lower 16 bits.
   */
  
  sum = 0;
  while (nbytes > 1)  {
    sum += *ptr++;
    nbytes -= 2;
  }
  
  /* mop up an odd byte, if necessary */
  if (nbytes == 1) {
    oddbyte = 0;            /* make sure top half is zero */
    *((u_char *) &oddbyte) = *(u_char *)ptr;   /* one byte only */
    sum += oddbyte;
  }
  
  /*
   * Add back carry outs from top 16 bits to low 16 bits.
   */
  
  sum  = (sum >> 16) + (sum & 0xffff);    /* add high-16 to low-16 */
  sum += (sum >> 16);                     /* add carry */
  answer = ~sum;          /* ones-complement, then truncate to 16 bits */
  return(answer);
}

/*
** debugging function
*/
void printIcmp(ICMPHDR *icmppacket)
{
 fprintf(stderr, "the details of the icmp packet are\n ");
 fprintf(stderr, "Type %d Code %d id %d seq %d CheckSum %d\n ", icmppacket->Type, icmppacket->Code, icmppacket->ID, icmppacket->Seq,icmppacket->Checksum);

}

void printIp(IPHDR *ippacket)
{
 fprintf(stderr, "the details of the ip packet are\n ");
 fprintf(stderr, "version %d Header Length %d id %d Service%x Length %04x TTL %d protocol %x CheckSum %d srcIP %s DestIp %s\n", ippacket->VIHL>>4, ((ippacket->VIHL & 0x0f) * 4), ippacket->ID,ippacket->TOS, ippacket->TotLen, ippacket->TTL, ippacket->Protocol, ippacket->Checksum,inet_ntoa(ippacket->iaSrc.s_addr), inet_ntoa(ippacket->iaDst.s_addr));

 fprintf(stderr, " %x flags %x fragmentation offset %x \n ",ippacket->FlagOff, ((ippacket->FlagOff & 0xe000) >> 13),(ippacket->FlagOff & 0x1fff) );

}

/*
** send an icmp Echo Request
*/
int SendEchoRequest(int s)
{
  static ECHOREQUEST echoReq;
  int nRet;
  int datasize = sizeof(struct timeval)+packetsize;

  
  /* Fill in ICMP Echo request packet*/
  echoReq.icmpHdr.Type            = ICMP_ECHOREQ;
  echoReq.icmpHdr.Code            = 0;
  echoReq.icmpHdr.Checksum        = 0;
  echoReq.icmpHdr.ID              = ident;       /*  helps me to uniquely identify ICMP packet on return */
  echoReq.icmpHdr.Seq             = transmited++;

  
  /* sending some junk as a part of data */	
  for (nRet = 0; nRet < packetsize; nRet++)
    echoReq.cData[nRet] = ' '+nRet;
	
  /* storing the packet sending time */
  gettimeofday( &echoReq.dwTime , (struct timezone *) 0);
  
  /* calculate the packet checksum */
  echoReq.icmpHdr.Checksum = in_cksum((u_short *)&echoReq, datasize);
  
if (debug)
   printIcmp(&echoReq.icmpHdr);

  /* sending the ICMP packet */
  nRet  = sendto(s,&echoReq,datasize, 0, (struct sockaddr *) &dest, sizeof(dest));
  
  /* check if the packet is send succesfully */
  if (nRet <0 || nRet < datasize )
       perror("sendto()");
  
  return (nRet);
}

void sigAlarm(int sigtype)
{
  printf("Host Unreachable \n");
  exit(0);
}


/*
** create a connection with the host machine
*/

int createConnection(char *host)
{
  struct protoent *proto;
  struct hostent *he;
  int sockfd;
  
  bzero((char *) &dest, sizeof(dest));
  dest.sin_family = AF_INET;
  
  /*
  ** Lookup hosttable from the DNS.
  */
  he = gethostbyname(host);
  
  if (he) 
    memcpy((caddr_t)&dest.sin_addr.s_addr, he->h_addr, he->h_length);
  else
    {
      (void) fprintf(stderr, " unknown hostname %s \n", host);
      return 0;
    }
  
  
  if (dest.sin_addr.s_addr == -1)
    {
      (void) fprintf(stderr, " unknown hostname %s \n", host);
      return 0;
    }
  
  proto = getprotobyname("icmp");
  if (proto == NULL)
    { 
      (void) fprintf(stderr," protocol icmp unidentified\n");
      return 0;
    }
  
  sockfd = socket(AF_INET, SOCK_RAW, proto->p_proto);
  
  if (sockfd <0)
    {
      (void) fprintf(stderr," unable to create RAW socket \n");
      return 0;
     }
  
  return sockfd;
  
}

/*
** happy days are here again thumps up the refreshing cola 
** 
*/
int RecvEchoReply(int s, u_char *pTTL, u_short *pSEQ, struct timeval *pdwTime) 
{
  ECHOREPLY echoReply;
  int nRet;
  int nAddrLen = sizeof(struct sockaddr_in);
  struct sockaddr_in lpsaFrom;
  
  for (;;)
    {
   alarm(1); /* to handle connections that may fail or get blocked as we donot have sufficient data */
      nRet = recvfrom(s, &echoReply , sizeof(echoReply), 0, (struct sockaddr *) &lpsaFrom, &nAddrLen);
      
      alarm(0); /* Turn off the alarm */
      
      /* Check return value */
      if (nRet == 0) {
	perror("recvfrom()");
	return -1;
      }
	
      *pTTL = echoReply.ipHdr.TTL;
      
      *pSEQ = echoReply.echoRequest.icmpHdr.Seq;
	
      *pdwTime = echoReply.echoRequest.dwTime;
	
	
      if (echoReply.echoRequest.icmpHdr.ID != ident) continue; /* if packet is not for us we will wait for our packet ;-) */
	
if (debug) {
   printIp(&echoReply.ipHdr);
   printIcmp(&echoReply.echoRequest.icmpHdr);
 }
      break;
	
    }
  
  return 1;
}

/*
** check if some data was send to us hopefully our icmp packet
*/
int waitforreply(int sockfd)
{
  
  int maxfd;
  int err;
  
  struct timeval  tv;
  fd_set   readfds;
  
  FD_ZERO(&readfds);
  maxfd=sockfd+1;
  FD_SET(sockfd,&readfds);
  
  tv.tv_sec=10;
  tv.tv_usec=0;
  
  err= select(maxfd,&readfds,NULL,NULL,&tv) ;
  if (err == 0)  return 0;
  
 if(FD_ISSET(sockfd,&readfds))
   return 1;
 
}

/*
** calculate the time difference between 2 timeval structures
*/
void tvsub( struct timeval *out, struct timeval *in )
{
  if( (out->tv_usec -= in->tv_usec) < 0 )   {
    out->tv_sec--;
    out->tv_usec += 1000000;
  }
  out->tv_sec -= in->tv_sec;
}

/*
** print the statistics 
*/
void printstat(char *host)
{
  (void) fprintf(stderr,"\n --- %s statistics --- \n ",host);
  (void) fprintf(stderr," %d packets transmitted, %d packets received, %0.2f%% packet loss\n",transmited,received, (received == 0) ? 100.00: 100.00 -((received*100.00)/transmited));
  
  (void) fprintf(stderr,"round-trip min/avg/max = %0.2f/%0.2f/%0.2f ms ", (received ==0 )? 0.0 :tmin,tsum/8,tmax );
}


/*
** the central command 
*/
int pinger(char * host)
{
  
  int nRet;
  int nLoop;
  int sockfd;
  u_char cTTL;
  u_short cSEQ;
  struct timeval tp;
  struct timeval tv;
  float triptime;
  
  
  if ((sockfd = createConnection(host)) == 0 ) return 0;
  
   signal(SIGALRM,sigAlarm);
   
   printf("\nPinging %s [%s] with %d bytes of data:\n", host, inet_ntoa(dest.sin_addr.s_addr), packetsize);
   
   for (nLoop = 0; nLoop < packetcount; nLoop++)
     {
       
       (void) SendEchoRequest(sockfd);
       
       if ( waitforreply(sockfd) == 0)
	 {
	   (void) printf( "\n Request Timed out\n");
	   continue;
	 }
       
       if (RecvEchoReply(sockfd, &cTTL,&cSEQ,&tp) == 0)  return 0;
       
       received++;

       /* calculate Round Trip Time */
       gettimeofday(&tv , (struct timezone *) 0);
       tvsub(&tv,&tp);
       triptime =  tv.tv_sec*1000.00+(tv.tv_usec/1000.00);
       tsum = tsum+triptime;
       
       if( triptime < tmin)
	 tmin = triptime;
       
       if( triptime > tmax)
	 tmax = triptime;
       
       printf("\nReply from: %s: bytes=%d time=%0.2fms icmp_seq %d TTL=%d ", host, packetsize, triptime, cSEQ, cTTL);
       
     }
   putchar('\n');
 printstat(host);
 putchar('\n');
 
 return 1;
 
}

int main(int argc, char *argv[])
{
  int c;
  char *host=NULL;
  
  /* process command line args */
  while ((c=getopt(argc,argv,"dpsh")) != EOF)
    {
      switch(c)
        {
        case 'd':
          debug = 1;
          if (debug)
	    printf("turn on debug mode.\n");
          break;
	  
        case 's':
          packetsize = atoi(argv[optind]);
          if (debug)
            printf("packet size specified to be %d bytes.\n",packetsize);
          break;
	  
        case 'p':
          packetcount = atoi(argv[optind]);
          if (debug)
            printf("view info invoked.\n");
	  
          if (argc <= optind) {
            help_msg(argv[0]);
            break;
          }
          break;
	  
        case 'h':
          host = argv[optind];
          if (debug)
            printf("hostname : %s.\n",host);
          break;
	  
        default:
          fprintf (stderr, "unrecognized arg >%s< %d\n", optarg,optind);
          help_msg(argv[0]);
          return (1);
	  
        }
    }
  
  setlinebuf(stdout);
  
  if (packetsize > MAX_PACKET_SIZE)
  {
    (void) fprintf(stderr, "packet size cannot be greater than  %d\n",(MAX_PACKET_SIZE+sizeof(struct timeval)));
    return 1;
  }
  

  /*
  **  i was trying to compare my program results and ping command results simultaniously for different hosts
  **  i actually displayed the results of the ping command through my program
  **
  **  search on google helped to understand that icmp packets are send to all processes so if u are listening for an icmp
  **  packet better be careful. i am assigning my process id to icmp id field to distinguish my packets from others.
  */
 ident=getpid();
 
 if (debug)
   printf("pid = %d\n",ident);
 
 if ( argc > 2 )
   return pinger(host); 
 else
   if ( argc  == 2 )
     return pinger(argv[1]); 
 
 
 help_msg(argv[0]);
 
 return 1;
 
}


